;; Generated package description from cursory.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "cursory" "1.2.0.0.20250722.60609" "Manage cursor styles using presets" '((emacs "29.1")) :commit "892c3b81037ece0e1753ab058e3cfda93f985693" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "cursor") :url "https://github.com/protesilaos/cursory")
